"""Drivers for various modem vendors."""

from typing import Any

__all__ = []


def export(defn: Any) -> None:  # noqa: ANN401
    """Module-level export decorator."""
    globals()[defn.__name__] = defn
    __all__.append(defn.__name__)  # noqa: PYI056
    return defn


from modem_info.drivers.utils import (
    DOCSISModemDriverProtocol,
    HTTPDOCSISModemDriverProtocol,
    HTTPModemDriver,
    HTTPModemDriverProtocol,
    ModemDriverProtocol,
)
from modem_info.drivers import (
    hitron,
)



"""Tests for the Hitron CODA-45 modem driver."""

import os.path

from importlib import import_module
from json import loads
from pathlib import Path
from time import time_ns

from pydantic import BaseModel


BASE: Path = Path(os.path.realpath(__file__)).parent / Path("test_data")
MODULE: str = "modem_info.drivers.hitron.coda45.models"
MODELS: dict[str, Path] = {
    "SystemInfo": Path(BASE / "getSysInfo.json"),
    "DOCSISProvisioning": Path(BASE / "getCMInit.json"),
    "LinkStatus": Path(BASE / "getLinkStatus.json"),
    "DOCSISOverview": Path(BASE / "getCmDocsisWan.json"),
    "DOCSISDownstream": Path(BASE / "dsinfo.json"),
    "DOCSISDownstreamOFDM": Path(BASE / "dsofdminfo.json"),
    "DOCSISUpstream": Path(BASE / "usinfo.json"),
    "DOCSISUpstreamOFDM": Path(BASE / "usofdminfo.json"),
}


def import_model(model: str) -> type[BaseModel]:
    """Dynamically import a data model."""
    return getattr(import_module(MODULE), model)


def run_model_tests() -> dict[str, BaseModel]:
    """Run data model validation tests."""
    ts: int = time_ns()
    results: dict[str, BaseModel] = {}
    for model, path in MODELS.items():
        data = None
        with path.open() as f:
            data = loads(f.read())
        results[model] = import_model(model)(timestamp=ts, data=data)
    return results


if __name__ == "__main__":
    from rich.pretty import pprint

    for result in run_model_tests().values():
        pprint(result, expand_all=True)



"""Plot statistics from your modem."""

from pathlib import Path
from typing import TYPE_CHECKING

import click
from pandas import read_csv
from plotly.graph_objects import Figure, Scatter

from modem_info.__main__ import main


if TYPE_CHECKING:
    from pandas import DataFrame
    from rich.console import Console


class COLORS:
    """Plotly colour theme."""

    NONE: str = "rgba(255, 255, 255, 0)"
    BLACK: str = "rgb(0, 0, 0)"
    WHITE: str = "rgb(255, 255, 255)"
    FORESTGREEN: str = "rgb(34, 139, 34)"
    FORESTGREEN_TRANS: str = "rgba(34, 139, 34, 0.2)"
    ROYALBLUE: str = "rgb(65, 105, 225)"
    ROYALBLUE_TRANS: str = "rgba(65, 105, 225, 0.2)"
    DARKORCHID: str = "rgb(153, 50, 204)"
    DARKORCHID_TRANS: str = "rgba(153, 50, 204, 0.2)"
    CRIMSON: str = "rgb(220, 20, 60)"
    CRIMSON_TRANS: str = "rgba(220, 20, 60, 0.2)"
    DARKORANGE: str = "rgb(255, 140, 0)"
    DARKORANGE_TRANS: str = "rgba(255, 140, 0, 0.2)"


@main.command()  # type: ignore[has-type]
@click.argument("file", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.help_option("-h", "--help")
@click.pass_context
def plot(ctx: click.Context, file: Path) -> None:
    """Plot statistics from your modem."""
    stdout: Console = ctx.obj["stdout"]  # noqa: F841
    stderr: Console = ctx.obj["stdout"]  # noqa: F841

    with file.open() as f:
        df: DataFrame = read_csv(f)

    data = [
        # Downstream Signal Min/Max/Avg (dBmV)
        Scatter(
            x=df["timestamp"],
            y=df["down_signal_max"],
            line_color=COLORS.FORESTGREEN,
            showlegend=False,
            hovertemplate=None,
            name="↓ Signal Max. (dBmV)",
            legendgroup="dsignal",
        ),
        Scatter(
            x=df["timestamp"],
            y=df["down_signal_min"],
            fill="tonexty",
            fillcolor=COLORS.FORESTGREEN_TRANS,
            line_color=COLORS.FORESTGREEN,
            showlegend=False,
            hovertemplate=None,
            name="↓ Signal Min. (dBmV)",
            legendgroup="dsignal",
        ),
        Scatter(
            x=df["timestamp"],
            y=df["down_signal_mean"],
            line_color=COLORS.FORESTGREEN,
            hovertemplate=None,
            name="↓ Signal Avg. (dBmV)",
            legendgroup="dsignal",
        ),
        # Downstream SNR Min/Max/Avg (dB)
        Scatter(
            x=df["timestamp"],
            y=df["down_snr_max"],
            line_color=COLORS.ROYALBLUE,
            showlegend=False,
            hovertemplate=None,
            name="↓ SNR Max. (dB)",
            legendgroup="dsnr",
        ),
        Scatter(
            x=df["timestamp"],
            y=df["down_snr_min"],
            fill="tonexty",
            fillcolor=COLORS.ROYALBLUE_TRANS,
            line_color=COLORS.ROYALBLUE,
            showlegend=False,
            hovertemplate=None,
            name="↓ SNR Min. (dB)",
            legendgroup="dsnr",
        ),
        Scatter(
            x=df["timestamp"],
            y=df["down_snr_mean"],
            line_color=COLORS.ROYALBLUE,
            hovertemplate=None,
            name="↓ SNR Avg. (dB)",
            legendgroup="dsnr",
        ),
        # Upstream Signal Min/Max/Avg (dB)
        Scatter(
            x=df["timestamp"],
            y=df["up_signal"],
            line_color=COLORS.DARKORCHID,
            hovertemplate=None,
            name="↑ Signal (dBmV)",
        ),
        # Downstream Octets
        Scatter(
            x=df["timestamp"],
            y=df["down_uncorrectables"],
            yaxis="y2",
            stackgroup="one",
            line_color=COLORS.CRIMSON,
            hovertemplate=None,
            name="↓ Uncorrectable",
            legend="legend2",
        ),
        Scatter(
            x=df["timestamp"],
            y=df["down_correcteds"],
            yaxis="y2",
            stackgroup="one",
            line_color=COLORS.DARKORANGE,
            hovertemplate=None,
            name="↓ Corrected",
            legend="legend2",
        ),
        Scatter(
            x=df["timestamp"],
            y=df["down_octets"],
            yaxis="y2",
            line_color=COLORS.ROYALBLUE,
            hovertemplate=None,
            stackgroup="one",
            name="↓ Total",
            legend="legend2",
        ),
        # Downstream Octets Δ
        Scatter(
            x=df["timestamp"],
            y=df["down_uncorrectables"].diff().shift(-1),
            yaxis="y3",
            stackgroup="one",
            line_color=COLORS.CRIMSON,
            hovertemplate=None,
            name="↓ Uncorrectable Δ",
            legend="legend3",
        ),
        # Scatter(
        #     x=df["timestamp"],
        #     y=df["down_correcteds"].diff().shift(-1),
        #     yaxis="y3",
        #     stackgroup="one",
        #     line_color=COLORS.DARKORANGE,
        #     hovertemplate=None,
        #     name="↓ Corrected Δ",
        #     legend="legend3",
        # ),
        Scatter(
            x=df["timestamp"],
            y=df["down_octets"].diff().shift(-1),
            yaxis="y3",
            stackgroup="one",
            line_color=COLORS.ROYALBLUE,
            hovertemplate=None,
            name="↓ Total Δ",
            legend="legend3",
        ),
    ]

    figure = Figure(
        data=data,
        layout=dict(
            title="Modem Statistics",
            hovermode="x",
            hoversubplots="axis",
            hoverlabel_namelength=30,
            showlegend=True,
            grid=dict(
                subplots=[["xy"], ["xy2"], ["xy3"]],
                xaxes=["x"],
                yaxes=["y", "y2", "y3"],
                ygap=0.05,
            ),
            yaxis=dict(
                title_text="dB",
                range=[0.0, 50.0],
            ),
            yaxis2=dict(
                title_text="Octets Total",
            ),
            yaxis3=dict(
                title_text="Octets Δ",
                range=[0, int(1e5)],
            ),
            legend=dict(
                title="Signal Levels",
            ),
            legend2=dict(
                title="Downstream Octets Total",
                y=0.65,
                yanchor="top",
            ),
            legend3=dict(
                title="Downstream Octets Δ",
                y=0.3,
                yanchor="top",
            ),
        ),
    )
    figure.show()



"""Drivers for the modem vendor Hitron."""

from typing import Any

__all__ = []


def export(defn: Any) -> None:  # noqa: ANN401
    """Module-level export decorator."""
    globals()[defn.__name__] = defn
    __all__.append(defn.__name__)  # noqa: PYI056
    return defn


from modem_info.drivers.hitron import coda45



"""Get detailed information and statistics from your modem."""

from __future__ import annotations

from csv import DictWriter
from ipaddress import IPv4Address, IPv6Address, ip_address
from pathlib import Path
from time import sleep
from typing import Any, TYPE_CHECKING

import click

from modem_info.__main__ import main
from modem_info.drivers.hitron.coda45 import HitronCoda45


if TYPE_CHECKING:
    from rich.console import Console
    from modem_info.drivers import HTTPDOCSISModemDriverProtocol


CSV_HEADERS = (
    "timestamp",
    "down_signal_min",
    "down_signal_max",
    "down_signal_mean",
    "down_signal_med",
    "down_snr_min",
    "down_snr_max",
    "down_snr_mean",
    "down_snr_med",
    "down_plc_power",
    "down_octets",
    "down_correcteds",
    "down_uncorrectables",
    "up_signal",
)


class IPAddressParamType(click.ParamType):
    """Represents a valid IPv4 or IPv6 address."""

    name = "ip address"

    def convert(
        self,
        value: Any,  # noqa: ANN401
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> IPv4Address | IPv6Address:
        """Convert the value to an IPv4/IPv6Address object."""
        try:
            return ip_address(value)
        except (TypeError, ValueError):
            self.fail(f"{value!r} is not a valid IP address", param, ctx)


def write_csv_docsis(modem: HitronCoda45, path: Path) -> None:
    """Write flattened statistics from your DOCSIS modem to CSV."""
    data = modem.docsis_statistics_flattened
    with Path(path / f"{modem.address}.csv").open("a") as f:
        writer = DictWriter(f, fieldnames=data.keys(), lineterminator="\n")
        if f.tell() == 0:
            writer.writeheader()
        writer.writerow(data)
        f.flush()


def write_jsonl_docsis(modem: HTTPDOCSISModemDriverProtocol, path: Path) -> None:
    """Write detailed statistics from your DOCSIS modem to JSONL."""
    models = {
        Path(path / f"{modem.address}_system_info.jsonl"): modem.system_info,
        Path(path / f"{modem.address}_link_status.jsonl"): modem.link_status,
        Path(path / f"{modem.address}_docsis_statistics.jsonl"): modem.docsis_statistics,
    }
    for file, model in models.items():
        with file.open("a") as f:
            f.write(model.model_dump_json() + "\n")
            f.flush()


@main.command()  # type: ignore[has-type]
@click.argument("address", type=IPAddressParamType())
@click.option(
    "-p",
    "--path",
    type=click.Path(exists=True, file_okay=False, writable=True, path_type=Path),
    default=Path("data"),
    show_default=True,
    help="Path to output files.",
)
@click.option(
    "-i",
    "--interval",
    type=click.FloatRange(min=5.0),
    default=60.0,
    show_default=True,
    help="Interval to record data on.",
)
@click.option(
    "-c",
    "--csv",
    is_flag=True,
    default=False,
    show_default=True,
    help="Write DOCSIS statistics to CSV.",
)
@click.option(
    "-j",
    "--json",
    is_flag=True,
    default=False,
    show_default=True,
    help="Write DOCSIS statistics to JSONL.",
)
@click.help_option("-h", "--help")
@click.pass_context
def get(  # noqa: PLR0913
    ctx: click.Context,
    address: IPv4Address | IPv6Address,
    path: Path,
    interval: float = 60.0,
    csv: bool = False,  # noqa: FBT001 FBT002
    json: bool = False,  # noqa: FBT001 FBT002
) -> int:
    """Get detailed information and statistics from a modem at ADDRESS."""
    stdout: Console = ctx.obj["stdout"]
    stderr: Console = ctx.obj["stdout"]

    if True not in (csv, json):
        stderr.log("[red bold]ERROR:[/red bold] Must select at least one output format: --csv, --json")

    try:
        modem = HitronCoda45(address)
    except BaseException:  # noqa: BLE001
        stderr.print_exception()
        stderr.log(f"[red bold]ERROR:[/red bold] Unable to connect to modem at {address}")
        return 1

    if csv:
        stdout.log(f"Beginning CSV output to {path}")
    if json:
        stdout.log(f"Beginning JSONL output to {path}")

    try:
        with stdout.status("Writing to file(s)..."):
            while True:
                if csv:
                    write_csv_docsis(modem, path)
                if json:
                    write_jsonl_docsis(modem, path)
                sleep(interval)
    except KeyboardInterrupt:
        stderr.log("Aborted by user")
    except BaseException:  # noqa: BLE001
        stderr.print_exception()
        stderr.log("[red bold]ERROR:[/red bold] Unknown exception")
        return 1

    return 0



"""Driver for the Hitron CODA-45."""

from __future__ import annotations

from datetime import datetime, timezone
from statistics import mean
from time import time, time_ns
from typing import Any, TypeVar, TYPE_CHECKING

from pydantic import BaseModel

from modem_info.drivers import HTTPModemDriver
from modem_info.drivers.hitron.coda45 import export, models


if TYPE_CHECKING:
    from httpx import Response

U = TypeVar("U", bound=BaseModel)


@export
class HitronCoda45(HTTPModemDriver):
    """Driver for the Hitron CODA-45."""

    _base = "/data"

    @staticmethod
    def _params(params: dict[str, str] | None = None) -> dict[str, str]:
        """Build the HTTP request parameters."""
        if params is None:
            params = {}
        return params | {"_": f"{time():.3f}".replace(".", "")}

    def _get(self, path: str, model: type[U]) -> U:
        """Request info from the modem and validate it against a data model."""
        ts = time_ns()
        response: Response = self._client.get(path, params=self._params())
        return model(timestamp=ts, data=response.json())

    @property
    def system_info(self) -> models.SystemInfo:
        """Get basic system information from the modem."""
        return self._get(self._base + "/getSysInfo.asp", models.SystemInfo)

    @property
    def link_status(self) -> models.LinkStatus:
        """Get link status from the modem."""
        return self._get(self._base + "/getLinkStatus.asp", models.LinkStatus)

    @property
    def docsis_provisioning(self) -> models.DOCSISProvisioning:
        """Get DOCSIS provisioning status from the modem."""
        return self._get(self._base + "/getCMInit.asp", models.DOCSISProvisioning)

    @property
    def docsis_overview(self) -> models.DOCSISOverview:
        """Get DOCSIS overview information from the modem."""
        return self._get(self._base + "/getCmDocsisWan.asp", models.DOCSISOverview)

    @property
    def docsis_downstream(self) -> models.DOCSISDownstream:
        """Get DOCSIS downstream information from the modem."""
        return self._get(self._base + "/dsinfo.asp", models.DOCSISDownstream)

    @property
    def docsis_downstream_flattened(self) -> models.DOCSISDownstreamFlattened:
        """Get flattened DOCSIS downstream information from the modem."""
        data = self.docsis_downstream
        signal_strengths = [channel.signal_strength for channel in data.data if channel.signal_strength is not None]
        snrs = [channel.snr for channel in data.data if channel.snr is not None]
        octets = [channel.octets for channel in data.data if channel.octets is not None]
        correcteds = [channel.corrected for channel in data.data if channel.corrected is not None]
        uncorrecteds = [channel.uncorrected for channel in data.data if channel.uncorrected is not None]

        return models.DOCSISDownstreamFlattened(
            timestamp=data.timestamp,
            num_channels=len(signal_strengths),
            signal_strength_min=min(signal_strengths),
            signal_strength_mean=mean(signal_strengths),
            signal_strength_max=max(signal_strengths),
            snr_min=min(snrs),
            snr_mean=mean(snrs),
            snr_max=max(snrs),
            octets_total=sum(octets),
            corrected_total=sum(correcteds),
            corrected_min=min(correcteds),
            corrected_mean=int(mean(correcteds)),
            corrected_max=max(correcteds),
            uncorrected_total=sum(uncorrecteds),
            uncorrected_min=min(uncorrecteds),
            uncorrected_mean=int(mean(uncorrecteds)),
            uncorrected_max=max(uncorrecteds),
        )

    @property
    def docsis_downstream_ofdm(self) -> models.DOCSISDownstreamOFDM:
        """Get DOCSIS downstream OFDM information from the modem."""
        return self._get(self._base + "/dsofdminfo.asp", models.DOCSISDownstreamOFDM)

    @property
    def docsis_downstream_ofdm_flattened(self) -> models.DOCSISDownstreamOFDMFlattened | None:
        """Get flattened DOCSIS downstream OFDM information from the modem."""
        data = self.docsis_downstream_ofdm
        num_channels = len([channel for channel in data.data if channel.plc_lock])
        if not num_channels:
            return None
        plc_powers = [channel.plc_power for channel in data.data if channel.plc_lock and channel.plc_power is not None]
        snrs = [channel.snr for channel in data.data if channel.plc_lock and channel.snr is not None]
        octets = [channel.octets for channel in data.data if channel.plc_lock and channel.octets is not None]
        correcteds = [channel.corrected for channel in data.data if channel.plc_lock and channel.corrected is not None]
        uncorrecteds = [
            channel.uncorrected for channel in data.data if channel.plc_lock and channel.uncorrected is not None
        ]

        return models.DOCSISDownstreamOFDMFlattened(
            timestamp=data.timestamp,
            num_channels=num_channels,
            plc_power_min=min(plc_powers),
            plc_power_mean=mean(plc_powers),
            plc_power_max=max(plc_powers),
            snr_min=min(snrs),
            snr_mean=mean(snrs),
            snr_max=max(snrs),
            octets_total=sum(octets),
            corrected_total=sum(correcteds),
            corrected_min=min(correcteds),
            corrected_mean=int(mean(correcteds)),
            corrected_max=max(correcteds),
            uncorrected_total=sum(uncorrecteds),
            uncorrected_min=min(uncorrecteds),
            uncorrected_mean=int(mean(uncorrecteds)),
            uncorrected_max=max(uncorrecteds),
        )

    @property
    def docsis_upstream(self) -> models.DOCSISUpstream:
        """Get DOCSIS upstream information from the modem."""
        return self._get(self._base + "/usinfo.asp", models.DOCSISUpstream)

    @property
    def docsis_upstream_flattened(self) -> models.DOCSISUpstreamFlattened:
        """Get flattened DOCSIS upstream information from the modem."""
        data = self.docsis_upstream
        signal_strengths = [channel.signal_strength for channel in data.data if channel.signal_strength is not None]

        return models.DOCSISUpstreamFlattened(
            timestamp=data.timestamp,
            num_channels=len(signal_strengths),
            signal_strength_min=min(signal_strengths),
            signal_strength_mean=mean(signal_strengths),
            signal_strength_max=max(signal_strengths),
        )

    @property
    def docsis_upstream_ofdm(self) -> models.DOCSISUpstreamOFDM:
        """Get DOCSIS upstream OFDM information from the modem."""
        return self._get(self._base + "/usofdminfo.asp", models.DOCSISUpstreamOFDM)

    @property
    def docsis_upstream_ofdm_flattened(self) -> models.DOCSISUpstreamOFDMFlattened | None:
        """Get flattened DOCSIS upstream OFDM information from the modem."""
        data = self.docsis_upstream_ofdm
        num_channels = len([channel for channel in data.data if channel.state])
        if not num_channels:
            return None
        line_digital_attenuations = [
            channel.line_digital_attenuation
            for channel in data.data
            if channel.state and channel.line_digital_attenuation is not None
        ]
        digital_attenuations = [
            channel.digital_attenuation
            for channel in data.data
            if channel.state and channel.digital_attenuation is not None
        ]
        report_powers = [
            channel.report_power for channel in data.data if channel.state and channel.report_power is not None
        ]
        report_power1_6s = [
            channel.report_power1_6 for channel in data.data if channel.state and channel.report_power1_6 is not None
        ]

        return models.DOCSISUpstreamOFDMFlattened(
            timestamp=data.timestamp,
            num_channels=num_channels,
            line_digital_attenuation_min=min(line_digital_attenuations),
            line_digital_attenuation_mean=mean(line_digital_attenuations),
            line_digital_attenuation_max=max(line_digital_attenuations),
            digital_attenuation_min=min(digital_attenuations),
            digital_attenuation_mean=mean(digital_attenuations),
            digital_attenuation_max=max(digital_attenuations),
            report_power_min=min(report_powers),
            report_power_mean=mean(report_powers),
            report_power_max=max(report_powers),
            report_power1_6_min=min(report_power1_6s),
            report_power1_6_mean=mean(report_power1_6s),
            report_power1_6_max=max(report_power1_6s),
        )

    @property
    def docsis_events(self) -> None:
        """Get DOCSIS events from the modem."""
        raise NotImplementedError  # TODO: Implement when we have example data

    @property
    def docsis_statistics(self) -> models.DOCSISStatistics:
        """Get DOCSIS statistics from the modem."""
        return models.DOCSISStatistics(
            timestamp=time_ns(),
            docsis_provisioning=self.docsis_provisioning.data[0],
            docsis_overview=self.docsis_overview.data[0],
            docsis_downstream=self.docsis_downstream.data,
            docsis_downstream_ofdm=self.docsis_downstream_ofdm.data,
            docsis_upstream=self.docsis_upstream.data,
            docsis_upstream_ofdm=self.docsis_upstream_ofdm.data,
        )

    @property
    def docsis_statistics_flattened(self) -> dict[str, Any]:
        """Get flattened DOCSIS statistics from the modem."""
        downstream = self.docsis_downstream_flattened
        downstream_ofdm = self.docsis_downstream_ofdm_flattened
        upstream = self.docsis_upstream_flattened

        if downstream_ofdm is not None:
            plc_power_mean = downstream_ofdm.plc_power_mean
            octets_total = downstream_ofdm.octets_total
            corrected_total = downstream_ofdm.corrected_total
            uncorrected_total = downstream_ofdm.uncorrected_total
        else:
            plc_power_mean = 0.0
            octets_total = 0
            corrected_total = 0
            uncorrected_total = 0

        return {
            "timestamp": datetime.now(timezone.utc).astimezone().isoformat(timespec="seconds"),
            "down_signal_min": f"{downstream.signal_strength_min:.3f}",
            "down_signal_mean": f"{downstream.signal_strength_mean:.3f}",
            "down_signal_max": f"{downstream.signal_strength_max:.3f}",
            "down_snr_min": f"{downstream.snr_min:.3f}",
            "down_snr_mean": f"{downstream.snr_mean:.3f}",
            "down_snr_max": f"{downstream.snr_max:.3f}",
            "down_plc_power": f"{plc_power_mean:.3f}",
            "down_octets_total": octets_total,
            "down_correcteds_total": corrected_total,
            "down_uncorrectables_total": uncorrected_total,
            "up_signal_mean": f"{upstream.signal_strength_mean:.3f}",
        }



"""Collect and plot detailed information and statistics from your modem."""

from datetime import datetime, timezone
from typing import TYPE_CHECKING

import click
from rich.console import Console

from modem_info import __name__, __version__, __copyright__


if TYPE_CHECKING:
    from collections.abc import Callable


_version: str = f"{__name__} v{__version__} -- {__copyright__}"


@click.group()
@click.help_option("-h", "--help")
@click.version_option(__version__, "-v", "--version", message=_version)
@click.pass_context
def main(ctx: click.Context) -> None:
    """
    Collect and plot detailed information and statistics from your modem.

    Run modem-info COMMAND --help for details on each command.
    """
    ctx.ensure_object(dict)

    log_time_format: str = "[%Y-%m-%dT%H:%M:%S.%f%z]"
    get_datetime: Callable = lambda: datetime.now(timezone.utc).astimezone()  # noqa: E731
    ctx.obj["stdout"] = Console(
        log_time_format=log_time_format,
        get_datetime=get_datetime,
    )
    ctx.obj["stderr"] = Console(
        log_time_format=log_time_format,
        get_datetime=get_datetime,
        stderr=True,
    )


if __name__ == "__main__":
    main(obj={})



"""Data models for the Hitron CODA-45 modem driver."""

from __future__ import annotations

from datetime import timedelta
from ipaddress import IPv4Address, IPv6Address, ip_address
from re import compile

from pydantic import BaseModel, Field, field_validator
from pydantic_extra_types.mac_address import MacAddress  # noqa: TCH002

from modem_info.drivers.hitron.coda45 import export


def normalize_str_to_bool(v: str) -> bool:
    """Normalize a string into a boolean."""
    return v.lower() in (
        "1",
        "true",
        "yes",
        "y",
        "on",
        "success",
        "permitted",
        "enabled",
        "enable",
    )


def normalize_str_to_str_or_none(v: str) -> str | None:
    """Normalize a string into a string or None."""
    if v.strip().lower() in ("", "na", "n/a"):
        return None
    return v


def normalize_str_to_int_or_zero(v: str) -> int:
    """Normalize a string into an integer or zero."""
    try:
        return int(v)
    except ValueError:
        return 0


def normalize_str_to_int_or_none(v: str) -> int | None:
    """Normalize a string into an integer or None."""
    try:
        return int(v)
    except ValueError:
        return None


def normalize_str_to_float_or_none(v: str) -> float | None:
    """Normalize a string into a float or None."""
    try:
        return float(v)
    except ValueError:
        return None


@export
class SystemInfoData(BaseModel):
    """Data element from /data/getSysInfo.asp."""

    hw_version: str = Field(validation_alias="hwVersion")
    sw_version: str = Field(validation_alias="swVersion")
    serial: str = Field(validation_alias="serialNumber")
    rf_mac: MacAddress = Field(validation_alias="rfMac")
    system_uptime: str = Field(validation_alias="systemUptime")
    system_time: str = Field(validation_alias="systemTime")


@export
class SystemInfo(BaseModel):
    """Data from /data/getSysInfo.asp."""

    timestamp: int
    data: list[SystemInfoData]


@export
class DOCSISProvisioningData(BaseModel):
    """Data element from /data/getCMInit.asp."""

    hw_init: bool = Field(validation_alias="hwInit")
    find_downstream: bool = Field(validation_alias="findDownstream")
    ranging: bool
    dhcp: bool
    time_of_day: bool = Field(validation_alias="timeOfday")
    download_config: bool = Field(validation_alias="downloadCfg")
    registration: bool
    eae_status: bool = Field(validation_alias="eaeStatus")
    bpi_status: str = Field(validation_alias="bpiStatus")
    network_access: bool = Field(validation_alias="networkAccess")
    traffic_status: bool = Field(validation_alias="trafficStatus")

    # Validators
    _normalize_str_to_bool = field_validator(
        "hw_init",
        "find_downstream",
        "ranging",
        "dhcp",
        "time_of_day",
        "download_config",
        "registration",
        "eae_status",
        "network_access",
        "traffic_status",
        mode="before",
    )(normalize_str_to_bool)


@export
class DOCSISProvisioning(BaseModel):
    """Data from /data/getCMInit.asp."""

    timestamp: int
    data: list[DOCSISProvisioningData]


@export
class LinkStatusData(BaseModel):
    """Data element from /data/getLinkStatus.asp."""

    status: bool = Field(validation_alias="LinkStatus")
    speed: str | None = Field(validation_alias="LinkSpeed")
    duplex: str | None = Field(validation_alias="LinkDuplex")

    # Validators
    @field_validator("status", mode="before")
    @classmethod
    def _normalize_link_status(cls, v: str) -> bool:
        return v.lower() == "up"

    @field_validator("speed", "duplex", mode="before")
    @classmethod
    def _normalize_str_or_none(cls, v: str) -> str | None:
        return None if v == "-" else v


@export
class LinkStatus(BaseModel):
    """Data from /data/getLinkStatus.asp."""

    timestamp: int
    data: list[LinkStatusData]


@export
class DOCSISOverviewData(BaseModel):
    """Data element from /data/getCmDocsisWan.asp."""

    config_name: str = Field(validation_alias="Configname")
    config_name_display: str = Field(validation_alias="ConfignameDisplay")
    network_access: bool = Field(validation_alias="NetworkAccess")
    ip_address: IPv4Address | IPv6Address | None = Field(validation_alias="CmIpAddress")
    netmask: IPv4Address | IPv6Address | None = Field(validation_alias="CmNetMask")
    gateway: IPv4Address | IPv6Address | None = Field(validation_alias="CmGateway")
    lease_duration: timedelta = Field(validation_alias="CmIpLeaseDuration")

    # Validators
    _normalize_str_to_bool = field_validator(
        "network_access",
        mode="before",
    )(normalize_str_to_bool)

    @field_validator("ip_address", "netmask", "gateway", mode="before")
    @classmethod
    def _normalize_ip_address(cls, v: str) -> IPv4Address | IPv6Address | None:
        try:
            return ip_address(v)
        except (TypeError, ValueError):
            return None

    @field_validator("lease_duration", mode="before")
    @classmethod
    def _normalize_lease_duration(cls, v: str) -> timedelta:
        re_lease_duration = compile(
            r"D: (?P<days>[0-9-]+) "
            r"H: (?P<hours>[0-9-]+) "
            r"M: (?P<minutes>[0-9-]+) "
            r"S: (?P<seconds>[0-9-]+)"  # noqa: COM812
        )
        match = re_lease_duration.fullmatch(v)
        if not match:
            return timedelta()
        return timedelta(
            days=normalize_str_to_int_or_zero(match["days"]),
            hours=normalize_str_to_int_or_zero(match["hours"]),
            minutes=normalize_str_to_int_or_zero(match["minutes"]),
            seconds=normalize_str_to_int_or_zero(match["seconds"]),
        )


@export
class DOCSISOverview(BaseModel):
    """Data from /data/getCmDocsisWan.asp."""

    timestamp: int
    data: list[DOCSISOverviewData]


@export
class DOCSISDownstreamData(BaseModel):
    """Data element from /data/dsinfo.asp."""

    port_id: int = Field(validation_alias="portId")
    frequency: int | None
    modulation: int | None
    signal_strength: float | None = Field(validation_alias="signalStrength")
    snr: float | None
    octets: int | None = Field(validation_alias="dsoctets")
    corrected: int | None = Field(validation_alias="correcteds")
    uncorrected: int | None = Field(validation_alias="uncorrect")
    channel_id: int | None = Field(validation_alias="channelId")

    # Validators
    _normalize_str_to_int_or_none = field_validator(
        "frequency",
        "modulation",
        "octets",
        "corrected",
        "uncorrected",
        "channel_id",
        mode="before",
    )(normalize_str_to_int_or_none)
    _normalize_str_to_float_or_none = field_validator(
        "signal_strength",
        "snr",
        mode="before",
    )(normalize_str_to_float_or_none)

    @field_validator("octets", mode="before")
    @classmethod
    def _normalize_bigint(cls, v: str) -> int | None:
        if " " not in v:
            try:
                return int(v)
            except ValueError:
                return None
        parts = v.split(" ")
        result = int(float(parts.pop(0)))
        while len(parts) > 0:
            op, val = parts.pop(0), int(float(parts.pop(0)))
            if op == "+":
                result += val
            if op == "*":
                result *= val
        return result


@export
class DOCSISDownstream(BaseModel):
    """Data from /data/dsinfo.asp."""

    timestamp: int
    data: list[DOCSISDownstreamData]


@export
class DOCSISDownstreamFlattened(BaseModel):
    """Flattened data from /data/dsinfo.asp."""

    timestamp: int
    num_channels: int
    signal_strength_min: float
    signal_strength_mean: float
    signal_strength_max: float
    snr_min: float
    snr_mean: float
    snr_max: float
    octets_total: int
    corrected_total: int
    corrected_min: int
    corrected_mean: int
    corrected_max: int
    uncorrected_total: int
    uncorrected_min: int
    uncorrected_mean: int
    uncorrected_max: int


@export
class DOCSISDownstreamOFDMData(BaseModel):
    """Data element from /data/dsofdminfo.asp."""

    receiver: int = Field(validation_alias="receive")
    fft_type: str | None = Field(validation_alias="ffttype")
    subcarrier_0_frequency: int | None = Field(validation_alias="Subcarr0freqFreq")
    plc_lock: bool = Field(validation_alias="plclock")
    ncp_lock: bool = Field(validation_alias="ncplock")
    mdc1_lock: bool = Field(validation_alias="mdc1lock")
    plc_power: float | None = Field(validation_alias="plcpower")
    snr: float | None = Field(validation_alias="SNR")
    octets: int | None = Field(validation_alias="dsoctets")
    corrected: int | None = Field(validation_alias="correcteds")
    uncorrected: int | None = Field(validation_alias="uncorrect")

    # Validators
    _normalize_str_to_str_or_none = field_validator(
        "fft_type",
        mode="before",
    )(normalize_str_to_str_or_none)
    _normalize_str_to_float_or_none = field_validator(
        "subcarrier_0_frequency",
        "plc_power",
        "snr",
        mode="before",
    )(normalize_str_to_float_or_none)
    _normalize_str_to_bool = field_validator(
        "plc_lock",
        "ncp_lock",
        "mdc1_lock",
        mode="before",
    )(normalize_str_to_bool)
    _normalize_str_to_int_or_none = field_validator(
        "octets",
        "corrected",
        "uncorrected",
        mode="before",
    )(normalize_str_to_int_or_none)


@export
class DOCSISDownstreamOFDM(BaseModel):
    """Data from /data/dsofdminfo.asp."""

    timestamp: int
    data: list[DOCSISDownstreamOFDMData]


@export
class DOCSISDownstreamOFDMFlattened(BaseModel):
    """Flattened data from /data/dsofdminfo.asp."""

    timestamp: int
    num_channels: int
    plc_power_min: float
    plc_power_mean: float
    plc_power_max: float
    snr_min: float
    snr_mean: float
    snr_max: float
    octets_total: int
    corrected_total: int
    corrected_min: int
    corrected_mean: int
    corrected_max: int
    uncorrected_total: int
    uncorrected_min: int
    uncorrected_mean: int
    uncorrected_max: int


@export
class DOCSISUpstreamData(BaseModel):
    """Data element from /data/usinfo.asp."""

    port_id: int = Field(validation_alias="portId")
    frequency: int | None
    bandwidth: int | None
    modulation: str | None = Field(validation_alias="modtype")
    docsis_mode: str | None = Field(validation_alias="scdmaMode")
    signal_strength: float | None = Field(validation_alias="signalStrength")
    channel_id: int | None = Field(validation_alias="channelId")

    # Validators
    _normalize_str_to_str_or_none = field_validator(
        "modulation",
        "docsis_mode",
        mode="before",
    )(normalize_str_to_str_or_none)
    _normalize_str_to_int_or_none = field_validator(
        "frequency",
        "bandwidth",
        "channel_id",
        mode="before",
    )(normalize_str_to_int_or_none)
    _normalize_str_to_float_or_none = field_validator(
        "signal_strength",
        mode="before",
    )(normalize_str_to_float_or_none)


@export
class DOCSISUpstream(BaseModel):
    """Data  from /data/usinfo.asp."""

    timestamp: int
    data: list[DOCSISUpstreamData]


@export
class DOCSISUpstreamFlattened(BaseModel):
    """Flattened data from /data/usinfo.asp."""

    timestamp: int
    num_channels: int
    signal_strength_min: float
    signal_strength_mean: float
    signal_strength_max: float


@export
class DOCSISUpstreamOFDMData(BaseModel):
    """Data element from /data/usofdminfo.asp."""

    channel_id: int = Field(validation_alias="uschindex")
    state: bool
    subcarrier_0_frequency: int | None = Field(validation_alias="frequency")
    line_digital_attenuation: float | None = Field(validation_alias="digAtten")
    digital_attenuation: float | None = Field(validation_alias="digAttenBo")
    bandwidth: float | None = Field(validation_alias="channelBw")
    report_power: float | None = Field(validation_alias="repPower")
    report_power1_6: float | None = Field(validation_alias="repPower1_6")
    fft_size: str | None = Field(validation_alias="fftVal")

    # Validators
    _normalize_str_to_bool = field_validator(
        "state",
        mode="before",
    )(normalize_str_to_bool)
    _normalize_str_to_str_or_none = field_validator(
        "fft_size",
        mode="before",
    )(normalize_str_to_str_or_none)
    _normalize_str_to_int_or_none = field_validator(
        "subcarrier_0_frequency",
        mode="before",
    )(normalize_str_to_int_or_none)
    _normalize_str_to_float_or_none = field_validator(
        "line_digital_attenuation",
        "digital_attenuation",
        "bandwidth",
        "report_power",
        "report_power1_6",
        mode="before",
    )(normalize_str_to_float_or_none)


@export
class DOCSISUpstreamOFDM(BaseModel):
    """Data from /data/usofdminfo.asp."""

    timestamp: int
    data: list[DOCSISUpstreamOFDMData]


@export
class DOCSISUpstreamOFDMFlattened(BaseModel):
    """Flattened data from /data/usofdminfo.asp."""

    timestamp: int
    num_channels: int
    line_digital_attenuation_min: float
    line_digital_attenuation_mean: float
    line_digital_attenuation_max: float
    digital_attenuation_min: float
    digital_attenuation_mean: float
    digital_attenuation_max: float
    report_power_min: float
    report_power_mean: float
    report_power_max: float
    report_power1_6_min: float
    report_power1_6_mean: float
    report_power1_6_max: float


@export
class DOCSISStatistics(BaseModel):
    """All data."""

    timestamp: int
    docsis_provisioning: DOCSISProvisioningData
    docsis_overview: DOCSISOverviewData
    docsis_downstream: list[DOCSISDownstreamData]
    docsis_downstream_ofdm: list[DOCSISDownstreamOFDMData]
    docsis_upstream: list[DOCSISUpstreamData]
    docsis_upstream_ofdm: list[DOCSISUpstreamOFDMData]



"""Collect and plot detailed information and statistics from your modem."""

from typing import Any

__all__ = []


def export(defn: Any) -> None:  # noqa: ANN401
    """Module-level export decorator."""
    globals()[defn.__name__] = defn
    __all__.append(defn.__name__)  # noqa: PYI056
    return defn


__copyright__ = "Copyright (c) 2024 Ryan Kozak"
from modem_info._version import __version__
from modem_info.get import get
from modem_info.plot import plot
from modem_info import drivers



"""Abstract base classes and protocols for modem drivers."""

from __future__ import annotations

from typing import Any, Protocol, TYPE_CHECKING

from httpx import Client


if TYPE_CHECKING:
    from ipaddress import IPv4Address, IPv6Address
    from pydantic import BaseModel


SCHEMES = [
    "http",
    "https",
]


class ModemDriverProtocol(Protocol):
    """Protocol for a generic modem driver."""

    @property
    def system_info(self) -> BaseModel:
        """Get basic system information from the modem."""
        ...

    @property
    def link_status(self) -> BaseModel:
        """Get link status from the modem."""
        ...


class DOCSISModemDriverProtocol(ModemDriverProtocol, Protocol):
    """Protocol for a generic DOCSIS modem driver."""

    @property
    def docsis_provisioning(self) -> BaseModel:
        """Get DOCSIS provisioning status from the modem."""
        ...

    @property
    def docsis_overview(self) -> BaseModel:
        """Get DOCSIS overview information from the modem."""
        ...

    @property
    def docsis_downstream(self) -> BaseModel:
        """Get DOCSIS downstream information from the modem."""
        ...

    @property
    def docsis_downstream_flattened(self) -> BaseModel:
        """Get flattened DOCSIS downstream information from the modem."""
        ...

    @property
    def docsis_downstream_ofdm(self) -> BaseModel:
        """Get DOCSIS downstream OFDM information from the modem."""
        ...

    @property
    def docsis_downstream_ofdm_flattened(self) -> BaseModel | None:
        """Get flattened DOCSIS downstream OFDM information from the modem."""
        ...

    @property
    def docsis_upstream(self) -> BaseModel:
        """Get DOCSIS upstream information from the modem."""
        ...

    @property
    def docsis_upstream_flattened(self) -> BaseModel:
        """Get flattened DOCSIS upstream information from the modem."""
        ...

    @property
    def docsis_upstream_ofdm(self) -> BaseModel:
        """Get DOCSIS upstream OFDM information from the modem."""
        ...

    @property
    def docsis_upstream_ofdm_flattened(self) -> BaseModel | None:
        """Get flattened DOCSIS upstream OFDM information from the modem."""
        ...

    @property
    def docsis_events(self) -> BaseModel | None:
        """Get DOCSIS events from the modem."""
        ...

    @property
    def docsis_statistics(self) -> BaseModel:
        """Get DOCSIS statistics from the modem."""
        ...

    @property
    def docsis_statistics_flattened(self) -> dict[str, Any]:
        """Get flattened DOCSIS statistics from the modem."""
        ...


class HTTPModemDriverProtocol(ModemDriverProtocol, Protocol):
    """Protocol for a generic HTTP-based modem driver."""

    address: IPv4Address | IPv6Address
    scheme: str
    params: dict[str, str]


class HTTPDOCSISModemDriverProtocol(HTTPModemDriverProtocol, DOCSISModemDriverProtocol, Protocol):
    """Protocol for a generic HTTP-based DOCSIS modem driver."""


class HTTPModemDriver:
    """Base class for a generic HTTP-based modem driver."""

    def __init__(
        self,
        address: IPv4Address | IPv6Address,
        scheme: str = "http",
        params: dict[str, str] | None = None,
    ) -> None:
        """Initialize the HTTP client to connect to the modem."""
        if scheme not in SCHEMES:
            msg = f"{scheme:!r} is not a value scheme"
            raise ValueError(msg)
        self.address = address
        self.scheme = scheme
        if params is None:
            self.params = {}
        else:
            self.params = params
        self._client = Client(base_url=f"{scheme}://{address}", params=params)
